@@ -70,8 +70,10 @@ chunk.apply <- function(input, FUN, ..., CH.MERGE=rbind, CH.MAX.SIZE=33554432,
     }
 }
 
-chunk.tapply <- function(input, FUN, ..., sep='\t', CH.MERGE=rbind, CH.MAX.SIZE=33554432) {
-    if (!inherits(inherits, "ChunkReader"))
-        reader <- chunk.reader(input)
-    .Call(chunk_tapply, reader, CH.MAX.SIZE, CH.MERGE, sep, FUN, parent.frame(), .External(pass, ...))
+chunk.tapply <- function(input, FUN, ..., sep, CH.MERGE=rbind, CH.MAX.SIZE=33554432) {
+    if (!inherits(input, "ChunkReader")) {
+        if (missing(sep)) sep <- '\t'
+        input <- chunk.reader(input, sep=sep)
+    } else if (missing(sep)) sep <- NULL
+    .Call(chunk_tapply, input, CH.MAX.SIZE, CH.MERGE, sep, FUN, parent.frame(), .External(pass, ...))
 }
